import logging
from abc import ABC
from typing import List, Optional

from ..Auth import Auth
from ..models.Downloable import MediaTrackSelector
from ..models.Season import Season
from ..utils.parser import safe_get

logger = logging.getLogger("pydisney")
logger = logger.getChild("Hit")


class Hit(ABC):
    def __init__(self, data):
        self.id: Optional[str] = None
        self.title: Optional[str] = None
        self.brief_desc: Optional[str] = None
        self.medium_desc: Optional[str] = None
        self.full_desc: Optional[str] = None
        self.genres: Optional[List[str]] = None
        self.rating: Optional[str] = None
        self.rating: Optional[str] = None
        self.artwork = None
        self.startYear: Optional[str] = None

        self.__seasons: List[Season] = []
        self.__is_movie: Optional[bool] = None
        self.__endYear: Optional[str] = None
        self.__credits: Optional[dict] = None
        self.__release_date: Optional[str] = None
        self.__length: Optional[int] = None
        self.__resource_id: Optional[str] = None
        self._parse_data(data)

    def _parse_data(self, data):
        self.id = data["id"]
        self.title = data["visuals"]["title"]
        self.brief_desc = safe_get(data, ["visuals", "description", "brief"])
        self.medium_desc = safe_get(data, ["visuals", "description", "medium"])
        self.full_desc = safe_get(data, ["visuals", "description", "full"])
        self.genres = safe_get(data, ["visuals", "metastringParts", "genres", "values"], [])
        self.rating = safe_get(data, ["visuals", "metastringParts", "ratingInfo", "rating", "text"])
        self.startYear = safe_get(data, ["visuals", "metastringParts", "releaseYearRange", "startYear"])
        self.artwork = safe_get(data, ["visuals", "artwork"])

    @staticmethod
    def parse_hits(data):
        hits = []
        for element in data:
            hits.append(Hit(element))
        return hits

    @property
    def is_movie(self):
        if self.__is_movie is None:
            self._get_more_data()
        return self.__is_movie

    @property
    def length(self):
        if not self.is_movie:
            raise ValueError("Series has no length property")
        if self.__length is None:
            self._get_more_data()
        return self.__length

    @property
    def seasons(self):
        if self.is_movie:
            raise ValueError("Movie has no seasons property")
        if not self.__seasons:
            self._get_more_data()
        return self.__seasons

    @property
    def credits(self):
        if not self.__credits:
            self._get_more_data()
        return self.__credits

    @property
    def endYear(self):
        if not self.__endYear:
            self._get_more_data()
        return self.__endYear

    @property
    def resourceId(self):
        if not self.is_movie:
            raise ValueError("Series has no resourceId property")
        if not self.__resource_id:
            self._get_playback_info()
        return self.__resource_id

    def _get_playback_info(self):
        url = "https://disney.api.edge.bamgrid.com/explore/v1.9/deeplink?action=playback&refId=9a280e53-fcc0-4e17-a02c-b1f40913eb0b&refIdType=deeplinkId"
        res = Auth.make_request("GET", url)
        self.__resource_id = res["data"]["deeplink"]["actions"][0]["resourceId"]

    def _get_more_data(self) -> None:
        url = f"https://disney.api.edge.bamgrid.com/explore/v1.9/page/entity-{self.id}"
        res = Auth.make_request("GET", url)
        self.__is_movie = True
        for container in res["data"]["page"]["containers"]:
            if container["type"] == "episodes":
                self.__is_movie = False
                for index, season_data in enumerate(container["seasons"]):
                    self.__seasons.append(Season(season_data, index))

            elif container["type"] == "set":
                # ignoring suggested and extras for now
                pass

            elif container["type"] == "details":
                data = safe_get(container, ["visuals", "credits"])
                if data:
                    formatted_data = {}
                    for entry in data:
                        heading = entry['heading']
                        items = [item['displayText'] for item in entry['items']]
                        formatted_data[heading] = items
                    self.__credits = formatted_data
                self.__endYear = safe_get(container, ["visuals", "metastringParts", "releaseYearRange", "endYear"], ignoreError=True)
                self.__length = safe_get(container, ["visuals", "duration", "runtimeMs"])

    def _get_m3u8_url(self):
        playback_link = "https://disney.playback.edge.bamgrid.com/v7/playback/ctr-regular"
        data = {"playback": {"attributes": {}}, "playbackId": self.resourceId}

        headers = {
            "x-bamsdk-client-id": "disney-svod-9zmsiwec",
            "x-bamsdk-platform": "javascript/windows/chrome",
            "x-bamsdk-version": "32.6",
            "x-dss-edge-accept": "vnd.dss.edge+json; version=2",
            "x-dss-feature-filtering": "true",
            "x-application-version": "1.1.2"
        }

        res = Auth.make_request("POST", playback_link, data=data, headers=headers)
        return res['stream']['sources'][0]['complete']["url"]

    def get_tracks(self) -> MediaTrackSelector:
        url = self._get_m3u8_url()
        return MediaTrackSelector(self.title, url)

    def __str__(self):
        return f"[Title={self.title}]"

    def __repr__(self):
        return self.title



import threading
import requests


class APIConfig:
    region = None
    language = None
    session = requests.Session()
    token = None
    refresh = None
    sessionId = None
    default_path = "downloads"
    auth = None
    _instance = None
    _lock = threading.Lock()

    def __new__(cls):
        if cls._instance is None:
            with cls._lock:
                # Another thread could have created the instance
                # before we acquired the lock. So check that the
                # instance is still nonexistent.
                if not cls._instance:
                    cls._instance = super().__new__(cls)
        return cls._instance



import logging
from typing import Optional, List

from .Auth import Auth
from .Config import APIConfig
from .models.Account import Account
from .models.Hit import Hit
from .models.Profile import Profile
from .models.ProgramType import MovieType, SeriesType
from .utils.parser import parse_profile

# Create a custom logger
logger = logging.getLogger("pydisney")
logger.setLevel(logging.INFO)

# Create a console handler
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)  # Set handler level to DEBUG

# Define a formatter and attach it to the handler
formatter = logging.Formatter("%(levelname)s: %(name)s: %(message)s")
console_handler.setFormatter(formatter)

# Add the handler to the logger
logger.addHandler(console_handler)


class DisneyAPI:
    def __init__(self, email: str, password: str, force_login: bool = False):

        self._auth = Auth(email=email, password=password, force_login=force_login)
        self._auth.get_auth_token()

        self.device_id = None
        self.device_platform = None
        self.sessions_id = None

        self.account: Optional[Account] = None
        self._account_init()
        self._session_init()

        APIConfig.region = "en" if self.account is None else self.account.country
        APIConfig.language = APIConfig.region

    def search(self, query: str) -> List[Hit]:
        res = Auth.make_pagination_request("GET", f"https://disney.api.edge.bamgrid.com/explore/v1.7/search?query={query}")
        for page in res:
            items = page["data"]["page"]["containers"]
            if not items:
                return []
            items = items[0]["items"]

            search_items = []
            for data in items:
                search_items.append(Hit(data))
            return search_items

    def _get_set(self, set_id) -> List[Hit]:
        url = f"https://disney.api.edge.bamgrid.com/explore/v1.7/set/{set_id}"
        res = Auth.make_pagination_request("GET", url)
        hits = []
        for page in res:
            items = page["data"]["set"]["items"]
            hits.extend(Hit.parse_hits(items))
        return hits

    def get_originals(self) -> List[Hit]:
        return self._get_set("3e935bc6-4984-4c49-acd8-a12a2e91ff40")

    def get_movies(self, movie_type: MovieType) -> List[Hit]:
        hits = self._get_set(movie_type.value)
        for hit in hits:
            hit.__is_movie = False
        return hits

    def get_series(self, series_type: SeriesType) -> List[Hit]:
        hits = self._get_set(series_type.value)
        for hit in hits:
            hit.__is_movie = False
        return hits

    def set_download_path(self, path: str) -> None:
        APIConfig.default_path = path

    def get_profiles(self) -> List[Profile]:
        graphql_query = {
            "query": """
                query {
                    me {
                        account {
                            profiles {
                                ...profile
                            }
                        }
                    }
                }
                fragment profile on Profile {
                    id
                    name
                    attributes {
                        avatar {
                            id
                            userSelected
                        }
                        isDefault
                        kidsModeEnabled
                        languagePreferences {
                            appLanguage
                            playbackLanguage
                            subtitleLanguage
                            subtitlesEnabled
                        }
                    }

                }
            """,
            "variables": {}
        }
        res = Auth.make_request("POST", "https://disney.api.edge.bamgrid.com/v1/public/graphql", data=graphql_query)

        profiles = []
        for profile in res["data"]["me"]["account"]["profiles"]:
            profiles.append(parse_profile(profile))
        return profiles

    def get_active_profile(self) -> Profile:
        graphql_query = {
            "query": """
                query {
                    me {
                        account {
                            activeProfile {
                                ...profile
                            }
                        }
                    }
                }
                fragment profile on Profile {
                    id
                    name
                    attributes {
                        avatar {
                            id
                            userSelected
                        }
                        isDefault
                        kidsModeEnabled
                        languagePreferences {
                            playbackLanguage
                            subtitleLanguage
                            subtitlesEnabled
                            appLanguage
                        }
                    }
                }
            """,
            "variables": {}
        }
        res = Auth.make_request("POST", "https://disney.api.edge.bamgrid.com/v1/public/graphql", data=graphql_query)
        profile = res["data"]["me"]["account"]["activeProfile"]
        return parse_profile(profile)

    def set_active_profile(self, profile_id: str, pin: str = None) -> None:
        access_token, refresh_token = self._auth.set_active_profile(APIConfig.token, profile_id, pin)
        APIConfig.token = access_token
        APIConfig.refresh = refresh_token

    def _account_init(self):
        graphql_query = {
            "query": """
                query {
                    me {
                        account {
                            ...account
                        }
                    }
                }
                fragment account on Account {
                    id
                    attributes {
                        consentPreferences {
                            dataElements {
                                name
                                value
                            }
                        }
                        dssIdentityCreatedAt
                        email
                        emailVerified
                        userVerified
                    }
                }
            """,
            "variables": {}
        }

        res = Auth.make_request("POST", "https://disney.api.edge.bamgrid.com/v1/public/graphql", data=graphql_query)

        acc_json = res["data"]["me"]["account"]
        account_id = acc_json["id"]
        email = acc_json["attributes"]["email"]
        email_verified = acc_json["attributes"]["emailVerified"]
        created_at = acc_json["attributes"]["dssIdentityCreatedAt"]
        try:
            country = acc_json["attributes"]["consentPreferences"]["dataElements"][0]["value"]
        except IndexError:
            # country seems to be missing for newly create accounts
            country = "en"
            logger.warning("Couldn't set country, fallback to default: EN-gb")

        account = Account(account_id=account_id, email=email, created_at=created_at, country=country,
                          is_email_verified=email_verified)
        self.account = account

    def _session_init(self):
        graphql_query = {
            "query": """
                query {
                    me {
                        activeSession {
                            ...session
                        }
                    }
                }
                fragment session on Session {
                     device {
                        id
                     platform
                     }
                    sessionId
                }
            """,
            "variables": {}
        }
        res = Auth.make_request("POST", "https://disney.api.edge.bamgrid.com/v1/public/graphql", data=graphql_query)

        sess_json = res["data"]["me"]["activeSession"]
        APIConfig.sessionId = sess_json["sessionId"]
        self.device_id = sess_json["device"]["id"]
        self.device_platform = sess_json["device"]["platform"]

    def get_token(self) -> str:
        return APIConfig.token

    def set_log_level(self, level: int) -> None:
        logger.setLevel(level)
        console_handler.setLevel(level)



from typing import Optional

from ..Auth import Auth
from ..models.Downloable import MediaTrackSelector
from ..utils.parser import safe_get


class Episode:
    def __init__(self, data):
        self.artwork: Optional[dict] = None
        self.durationMs: Optional[int] = None
        self.brief_description: Optional[str] = None
        self.medium_description: Optional[str] = None
        self.full_description: Optional[str] = None
        self.episodeNumber: Optional[int] = None
        self.episodeTitle: Optional[str] = None
        self.fullEpisodeTitle: Optional[str] = None
        self.rating: Optional[dict] = None
        self.id: Optional[str] = None
        self.resourceId: Optional[str] = None

        self._parse_data(data)

    def _parse_data(self, data) -> None:
        self.id = safe_get(data, ["id"])
        self.resourceId = safe_get(data, ["actions", 0, "resourceId"])
        self.artwork = safe_get(data, ["visuals", "artwork"])
        self.durationMs = safe_get(data, ["visuals", "durationMs"])
        self.brief_description = safe_get(data, ["visuals", "description", "brief"])
        self.medium_description = safe_get(data, ["visuals", "description", "medium"])
        self.full_description = safe_get(data, ["visuals", "description", "full"])

        self.episodeNumber = safe_get(data, ["visuals", "episodeNumber"])
        self.episodeTitle = safe_get(data, ["visuals", "episodeTitle"])
        self.fullEpisodeTitle = safe_get(data, ["visuals", "fullEpisodeTitle"])

        self.rating = safe_get(data, ["visuals", "metastringParts", "ratingInfo", "rating", "text"])

    def __str__(self):
        return f"[Title={self.fullEpisodeTitle}, Episode={self.episodeNumber}]"

    def _get_m3u8_url(self):
        playback_link = "https://disney.playback.edge.bamgrid.com/v7/playback/ctr-regular"
        data = {"playback": {"attributes": {}}, "playbackId": self.resourceId}

        headers = {
            "x-bamsdk-client-id": "disney-svod-9zmsiwec",
            "x-bamsdk-platform": "javascript/windows/chrome",
            "x-bamsdk-version": "32.6",
            "x-dss-edge-accept": "vnd.dss.edge+json; version=2",
            "x-dss-feature-filtering": "true",
            "x-application-version": "1.1.2"
        }

        res = Auth.make_request("POST", playback_link, data=data, headers=headers)
        return res['stream']['sources'][0]['complete']["url"]

    def get_tracks(self) -> MediaTrackSelector:
        url = self._get_m3u8_url()
        return MediaTrackSelector(self.fullEpisodeTitle, url)

    def __repr__(self):
        return self.fullEpisodeTitle



import os
from collections import defaultdict
from dataclasses import dataclass
from typing import Optional, List
from urllib.parse import urljoin

import m3u8
import requests

from ..Config import APIConfig
from ..utils.helper import rename_filename


@dataclass
class Subtitle:
    title: str
    language: str
    uri: str
    name: str = ""
    group_id: str = ""
    forced: str = None

    def download(self, name=None):
        if not name:
            name = self.title + " - " + self.language + ".vtt"
        name = rename_filename(name)

        output_path = os.path.join(APIConfig.default_path, name)

        response = requests.get(self.uri)
        response.raise_for_status()

        lines = response.text.splitlines()
        vtt_segments = [
            urljoin(self.uri, line.strip())
            for line in lines
            if line and not line.startswith("#")
        ]

        with open(output_path, "wb") as outfile:
            for idx, seg_url in enumerate(vtt_segments):
                seg_resp = requests.get(seg_url)
                seg_resp.raise_for_status()
                content = seg_resp.content

                if idx != 0:
                    content = b"\n".join(content.splitlines()[1:])
                outfile.write(content + b"\n")

        return output_path

    def __str__(self):
        return f"Subtitle[language='{self.language}', name='{self.name}', forced={self.forced}, title='{self.title}']"

    def __repr__(self):
        return self.__str__()


@dataclass
class Audio:
    title: str
    language: str
    uri: str
    name: str = ""
    group_id: str = ""
    quality: str = ""
    codec: str = ""
    channels: str = ""

    def download(self, name=None):
        if not name:
            ext = self.codec
            if self.codec == "eac":
                ext = "ac3"
            name = self.title + " - " + self.language + "." + ext

        name = rename_filename(name)

        output_path = os.path.join(APIConfig.default_path, name)

        os.system(
            f'ffmpeg -protocol_whitelist file,http,https,tcp,tls,crypto -i "{self.uri}" -c copy "{output_path}" -preset ultrafast -loglevel warning -hide_banner -stats')

    def __str__(self):
        return f"Audio[language='{self.language}', name='{self.name}', title='{self.title}', quality={self.quality}, codec={self.codec}, channels={self.channels}]"

    def __repr__(self):
        return self.__str__()


class AudioTrackGroup:
    def __init__(self, tracks: List['Audio']):
        self._tracks = tracks

    def all(self):
        return self._tracks

    def filter_by_quality(self, quality: str):
        return AudioTrackGroup([t for t in self._tracks if t.quality == quality])

    def filter_by_codec(self, codec: str):
        return AudioTrackGroup([t for t in self._tracks if (t.codec or "").lower() == codec.lower()])

    def get_best_quality(self, preferred_codecs: Optional[List[str]] = None) -> Optional['Audio']:
        def parse_quality(q: str) -> int:
            try:
                return int(q.replace("k", "")) * 1000
            except (AttributeError, ValueError):
                return 0

        def codec_rank(codec: str):
            if preferred_codecs:
                try:
                    return preferred_codecs.index(codec.lower())
                except ValueError:
                    return len(preferred_codecs)
            return 0

        return max(
            self._tracks,
            key=lambda a: (
                parse_quality(a.quality),
                int(a.channels or 0),
                -codec_rank(a.codec)
            ),
            default=None
        )

    def first(self) -> Optional['Audio']:
        return self._tracks[0] if self._tracks else None

    def __str__(self):
        return f"AudioTrackGroup[length={len(self._tracks)}]"

    def __repr__(self):
        return self.__str__()

class SubtitleTrackGroup:
    def __init__(self, tracks: List['Subtitle']):
        self._tracks = tracks

    def all(self):
        return self._tracks

    def get_forced(self):
        return SubtitleTrackGroup([t for t in self._tracks if t.forced])

    def get_unforced(self):
        return SubtitleTrackGroup([t for t in self._tracks if not t.forced])

    def first(self) -> Optional['Subtitle']:
        return self._tracks[0] if self._tracks else None

    def __str__(self):
        return f"SubtitleTrackGroup[length={len(self._tracks)}]"

    def __repr__(self):
        return self.__str__()

class MediaTrackSelector:
    def __init__(self, title, m3u8_content):
        self.title = title
        self.subs, self.audios = self._parse_m3u8_from_url(m3u8_content)

    def _parse_m3u8_from_url(self, url):
        m3u8_obj = m3u8.load(url)
        subs = defaultdict(list)
        audios = defaultdict(list)

        for media in m3u8_obj.media:
            if not media.language:
                continue

            full_uri = urljoin(url, media.uri)
            lang_key = media.language.lower()
            if media.type == "SUBTITLES":
                subs[lang_key].append(
                    Subtitle(
                        title=self.title,
                        language=lang_key,
                        uri=full_uri,
                        name=media.name,
                        group_id=media.group_id,
                        forced=media.forced == "YES"
                    )
                )
            elif media.type == "AUDIO":
                group_id = media.group_id or ""
                parts = group_id.split("-")
                codec = parts[0] if len(parts) > 0 else ""
                quality = parts[1] if len(parts) > 1 else ""

                audios[lang_key].append(
                    Audio(
                        title=self.title,
                        language=lang_key,
                        uri=full_uri,
                        name=media.name or "",
                        group_id=group_id,
                        quality=quality,
                        codec=codec,
                        channels=media.channels
                    )
                )
        return subs, audios

    def get_audios(self, language="en") -> AudioTrackGroup:
        return AudioTrackGroup(self.audios.get(language.lower(), []))

    def get_subtitles(self, language="en") -> SubtitleTrackGroup:
        return SubtitleTrackGroup(self.subs.get(language.lower(), []))

    @property
    def all_audios(self):
        return list(self.audios.values())

    @property
    def all_subtitles(self):
        return list(self.subs.values())

    def available_languages(self):
        return {
            "subtitles": list(self.subs.keys()),
            "audio": list(self.audios.keys())
        }

    def __str__(self):
        return f"MediaTrackSelector[audio={len(self.audios)}, subtitle={len(self.subs)}]"



import json
import logging
import re
from datetime import datetime
from json import JSONDecodeError
from typing import List

import requests

from .Config import APIConfig
from .Exceptions import AuthException, ApiException
from .utils.helper import update_file

logger = logging.getLogger("pydisney")
logger = logger.getChild("Auth")


class Auth:
    def __init__(self, email, password, force_login):
        self._email = email
        self._force_login = force_login
        self._password = password
        APIConfig.auth = self

    def _obtain_client_api_key(self):
        """Step 1: Obtain client's api key by parsing html page..."""
        res = APIConfig.session.get("https://www.disneyplus.com")
        pattern = r'"clientId"\s*:\s*"([^"]+)"\s*,\s*"clientApiKey"\s*:\s*"([^"]+)"'
        match = re.search(pattern, res.text)

        if match:
            client_api_key = match.group(2)
            return client_api_key
        else:
            raise AuthException("Unable to obtain client api key. Please open an issue.")

    def _register_device(self, client_api_key):
        """Step 2: Register a device with previously obtained client_api_key"""
        graphql_query = {
            "operationName": "registerDevice",
            "query": "mutation registerDevice($input: RegisterDeviceInput!) { registerDevice(registerDevice: $input) { grant { grantType assertion }, activeSession { partnerName profile { id } } } }",
            "variables": {
                "input": {
                    "deviceFamily": "N/A",
                    "applicationRuntime": "N/A",
                    "deviceProfile": "N/A",
                    "deviceLanguage": "N/A",
                    "devicePlatformId": "N/A",
                    "attributes": {
                        "brand": "N/A",
                        "browserName": "N/A",
                        "browserVersion": "N/A",
                        "manufacturer": "N/A",
                        "operatingSystem": "N/A",
                        "operatingSystemVersion": "N/A",
                    }
                }
            }
        }
        res = APIConfig.session.post("https://disney.api.edge.bamgrid.com/graph/v1/device/graphql", json=graphql_query, headers={"Authorization": f"Bearer {client_api_key}"})
        json_data = res.json()
        access_token = json_data["extensions"]["sdk"]["token"]["accessToken"]
        return access_token

    def _login(self, access_token):
        """Step 3. Perform a login with access token obtained during register device"""
        graphql_query = {
            "query": "mutation login($input: LoginInput!) { login(login: $input) { actionGrant account { activeProfile { id } profiles { id attributes { isDefault parentalControls { isPinProtected } } } } activeSession { isSubscriber } identity { personalInfo { dateOfBirth gender } flows { personalInfo { requiresCollection eligibleForCollection } } } } }",
            "variables": {
                "input": {
                    "email": self._email,
                    "password": self._password
                }
            },
            "operationName": "login"
        }

        res = APIConfig.session.post("https://disney.api.edge.bamgrid.com/v1/public/graphql", json=graphql_query, headers={"Authorization": f"Bearer {access_token}"})
        json_data = res.json()
        profiles = json_data["data"]["login"]["account"]["profiles"]
        access_token = json_data["extensions"]["sdk"]["token"]["accessToken"]
        return access_token, profiles

    def set_active_profile(self, access_token, profile_id, pin=None):
        """Step 4. Set an active profile. Only access token returned here can be used to query data"""
        graphql_mutation = {
            "query": """mutation switchProfile($input: SwitchProfileInput!) { switchProfile(switchProfile: $input) { account { ...account } } } fragment account on Account { id }""",
            "variables": {
                "input": {
                    "profileId": profile_id,
                    "entryPin": pin
                }
            },
            "operationName": "switchProfile"
        }
        res = APIConfig.session.post("https://disney.api.edge.bamgrid.com/v1/public/graphql", json=graphql_mutation, headers={"Authorization": f"Bearer {access_token}"})
        json_data = res.json()
        access_token = json_data["extensions"]["sdk"]["token"]["accessToken"]
        refresh_token = json_data["extensions"]["sdk"]["token"]["refreshToken"]
        return access_token, refresh_token

    def _get_valid_profile_id(self, profiles):
        default_locked = False
        for profile in profiles:
            if profile["attributes"]["isDefault"]:
                if profile["attributes"]["parentalControls"]["isPinProtected"]:
                    default_locked = True
                    logger.warning("Default profile is locked! Defaulting to a different one")
            if default_locked and not profile["attributes"]["parentalControls"]["isPinProtected"]:
                return profile["id"]

        raise AuthException("Unable to pick a profile: all profiles are locked. Manually set a profile and supply a PIN.")

    def _get_auth_token_tru_api(self):
        logger.info("Loging using disney's api")
        logger.warning("Loging using Disney's api repeatedly will cause account blocks")
        client_api_key = self._obtain_client_api_key()
        device_access_token = self._register_device(client_api_key)
        non_profile_access_token, profiles = self._login(device_access_token)

        profile_id = self._get_valid_profile_id(profiles)
        access_token, refresh_token = self.set_active_profile(non_profile_access_token, profile_id)
        APIConfig.token = access_token
        APIConfig.refresh = refresh_token
        update_file()

    @staticmethod
    def refreshToken():
        with open("token.json", "r", encoding="utf8") as file:
            data = json.load(file)

            refresh = data["refresh"]

            logger.info("Refreshing access token using refresh token from token.json file")
            graph_mutation = {
                "query": "mutation refreshToken($input:RefreshTokenInput!){refreshToken(refreshToken:$input){activeSession{sessionId}}}",
                "variables": {
                    "input": {
                        "refreshToken": refresh
                    }
                },
                "operationName": "refreshToken"
            }

            res = APIConfig.session.post(url="https://disney.api.edge.bamgrid.com/graph/v1/device/graphql",
                                         json=graph_mutation,
                                         headers={"authorization": APIConfig.auth._obtain_client_api_key()})
            if not res.ok:
                raise AuthException(res)

            if not res.json()["data"]:
                errors = []
                for error in res.json()["errors"]:
                    errors.append(error)
                logger.info("Couldn't refresh access token with refresh token:\n%s", errors)
                #  trying one last time to get access token via standard login procedure using email and password
                APIConfig.auth._get_auth_token_tru_api()
                return

            res_json = res.json()

            APIConfig.token = res_json["extensions"]["sdk"]["token"]["accessToken"]
            APIConfig.refresh = res_json["extensions"]["sdk"]["token"]["refreshToken"]
            update_file()

    @staticmethod
    def make_request(method: str, url: str, data: dict = None, headers: dict = None, params: dict = None, files: dict = None) -> dict:
        default_headers = {
            'accept': 'application/vnd.media-service+json; version=6',
            'User-Agent': "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
            'x-bamsdk-platform': "macOS",
            'x-bamsdk-version': '23.1',
            'x-dss-edge-accept': 'vnd.dss.edge+json; version=2',
            'x-dss-feature-filtering': 'true',
            'Origin': 'https://www.disneyplus.com',
            'authorization': APIConfig.token
        }

        if params is None:
            params = {}

        if headers is None:
            headers = {}

        logger.debug(f"Calling... Url={url}, Method={method}, Headers={headers}")

        headers.update(default_headers)
        response = requests.request(method, url, headers=headers, json=data, params=params, files=files)

        if not response.ok:
            raise ApiException(response)

        if response.status_code == 200:
            return response.json()

    @staticmethod
    def make_pagination_request(method: str, url: str, data: dict = None, headers: dict = None, params: dict = None, files: dict = None) -> List[dict]:
        if params is None:
            params = {}
        params.setdefault('limit', 200)
        params.setdefault('offset', 0)

        aggregated_results = []
        current_offset = params['offset']
        limit = params['limit']

        while True:
            logger.debug(f"Fetching page with offset={current_offset}, limit={limit}")
            params['offset'] = current_offset

            # Use the existing make_request method
            response_json = Auth.make_request(
                method=method,
                url=url,
                data=data,
                headers=headers,
                params=params,
                files=files
            )

            from .utils.parser import find_pagination_values  # circular import
            pagination_values = find_pagination_values(response_json)

            logger.debug(f"Pagination Values: {pagination_values}")

            aggregated_results.append(response_json)

            # Check if there are more pages
            if pagination_values and pagination_values.get('hasMore', False):
                current_offset += limit
            else:
                break

        return aggregated_results

    @staticmethod
    def make_stream_request(url):
        headers = {
            'User-Agent': "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36"}

        res = APIConfig.session.get(url=url, headers=headers, stream=True, timeout=10)
        if res.status_code == 401:
            raise AuthException(res)
        if not res.ok:
            raise ApiException(res)

        return res

    def get_auth_token(self):
        if self._force_login:
            self._get_auth_token_tru_api()
            return
        try:
            with open("token.json", "r", encoding="utf8") as file:

                data = json.load(file)
                token = data["token"]
                refresh = data["refresh"]
                expiration_time_str = data["expiration_time"]

                # Convert the expiration time string back to a datetime object
                expiration_time = datetime.strptime(expiration_time_str, "%Y-%m-%d %H:%M:%S")

                # Check if the token is still valid
                current_time = datetime.now()
                if current_time < expiration_time:
                    logger.info("Authenticating using token from token.json file")

                    APIConfig.token = token
                    APIConfig.refresh = refresh
                else:
                    logger.info("Access token is expired, refreshing...")

                    self.refreshToken()

        except (FileNotFoundError, KeyError, JSONDecodeError):

            logger.info("Creating token.json file")
            open("token.json", "a+", encoding="utf8").close()

            self._get_auth_token_tru_api()
